# -*- coding: utf-8 -*-
#
# Copyright 2022, Philippe Kueck <projects at unixadm dot org>
#
# This software may be freely redistributed under the terms of the GNU
# general public license version 2.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


import re
from time import strftime
from hashlib import sha256
import rpm
import pymysql

def rpminfo(filename, chksum=True):
    ts = rpm.TransactionSet()
    hsh = sha256()
    with open(filename, 'rb') as fh:
        hdr = ts.hdrFromFdno(fh)
        if chksum:
            fh.seek(0)
            for chunk in iter(lambda: fh.read(hsh.block_size), b''):
                hsh.update(chunk)
    if len(hdr[rpm.RPMTAG_CHANGELOGTEXT]) > 0:
        cl = hdr[rpm.RPMTAG_CHANGELOGTEXT][0]
    else: cl = None

    return {
        'master': re.sub(
            '.%s.%s.src.rpm$' % (
                hdr[rpm.RPMTAG_VERSION], hdr[rpm.RPMTAG_RELEASE]
            ), '', hdr[rpm.RPMTAG_SOURCERPM]),
        'srpm': hdr[rpm.RPMTAG_SOURCERPM],
        'name': hdr[rpm.RPMTAG_NAME],
        'epoch': str(hdr[rpm.RPMTAG_EPOCH] or 0),
        'version': hdr[rpm.RPMTAG_VERSION],
        'release': hdr[rpm.RPMTAG_RELEASE],
        'arch': hdr[rpm.RPMTAG_ARCH],
        'changelog': cl,
        'buildtime': hdr[rpm.RPMTAG_BUILDTIME],
        'checksum': hsh.hexdigest()
    }

issue_types = [
    'recommended', 'optional', 'feature', 'enhancement',
    'newpackage', 'bugfix', 'security'
]

class RepoIssueDB:

    def __init__(self, database, user, password, host="localhost"):
        self._db = pymysql.connect(
            host=host, user=user, password=password, database=database,
            autocommit=False
        )
        self.cursor = self._db.cursor(pymysql.cursors.DictCursor)
        self.cursor.execute(f"""
            select table_name from information_schema.tables where table_schema = '{database}'
        """)
        tables = [t['table_name'] for t in self.cursor.fetchall()]
        if "issues" not in tables:
            self.cursor.execute("""
                create table issues (
                    `issue_id` int(7) unsigned not null primary key,
                    `issued_date` timestamp default current_timestamp,
                    `type` enum(%s) not null default "optional",
                    `title` varchar(255),
                    `description` text,
                    `approved` char(1) not null default 0,
                    unique key (`issue_id`)
                ) engine=innodb
            """ % ",".join([f"'{t}'" for t in issue_types]))
        if "packages" not in tables:
            self.cursor.execute("""
                create table packages (
                    `package_id` int unsigned not null auto_increment primary key,
                    `issue_id` int(7) unsigned not null,
                    `name` varchar(255) not null,
                    `epoch` varchar(3) not null default 0,
                    `version` varchar(15) not null,
                    `release` varchar(15) not null,
                    `arch` enum('noarch', 'x86_64', 'i686') not null,
                    `srpm` varchar(255) not null,
                    `filename` varchar(255) not null,
                    `checksum` char(64),
                    foreign key (`issue_id`) references issues(`issue_id`) on
                        delete cascade on update cascade,
                    unique key (`name`, `epoch`, `version`, `release`, `arch`)
                ) engine=innodb
            """)
        if "meta" not in tables:
            self.cursor.execute("""
                create table meta (
                    `short` varchar(16) not null,
                    `releasever` int unsigned not null,
                    `name` varchar(16) not null,
                    `description` varchar(255) not null,
                    primary key(releasever, name)
                ) engine=innodb
            """)

        self._db.commit()


    def new_issue(self, rpms, title=None, issue_type=None):
        if len(rpms) == 0: raise ValueError("no rpms given")
        if issue_type == None: issue_type = "optional"

        description = None
        buildtime = 0
        query_values = []

        issue_min = int(strftime('%Y000'))
        self.cursor.execute("""
            select max(issue_id) as max from issues where issue_id > %s
        """, issue_min)
        res = self.cursor.fetchone()
        if res['max'] is None: issue_id = issue_min+1
        else: issue_id = res['max']+1
        self.cursor.execute("insert into `issues` (`issue_id`) values (%s)", issue_id)

        for rpm in rpms:
            rpm_data = rpminfo(rpm)
            query_values.append((
                issue_id, rpm_data['name'], rpm_data['epoch'], rpm_data['version'],
                rpm_data['release'], rpm_data['arch'], rpm_data['srpm'], rpm,
                rpm_data['checksum']
            ))
            if title is None: title = rpm_data['master']
            if description is None: description = rpm_data['changelog']
            if buildtime == 0: buildtime = rpm_data['buildtime']

        try:
            rows = self.cursor.executemany("""
                insert into `packages` (`issue_id`, `name`, `epoch`, `version`,
                    `release`, `arch`, `srpm`, `filename`, `checksum`)
                    values (%s, %s, %s, %s, %s, %s, %s, %s, %s)
            """, query_values)
        except pymysql.err.IntegrityError:
            raise AssertionError("duplicate packages")
        if len(rpms) != rows: raise AssertionError(f"len(rpms) != {rows}")

        detected_type = None
        self.cursor.execute(
            "select count(title) as cnt from issues where title = %s", title
        )
        res = self.cursor.fetchone()
        if res['cnt'] == 0:
            detected_type = "newpackage"
        elif description is not None:
            if re.search(r'C(?:VE|AN)[#-]', description):
                detected_type = "security"
            elif re.search(r'(?:BZ[#-]|(bug ?)?fix(?:es|d|\b))?', description):
                detected_type = "bugfix"
            else:
                detected_type = issue_type
        else:
            detected_type = issue_type
        self.cursor.execute("""
            update `issues` set `issued_date` = from_unixtime(%s), `description` = %s,
            `title` = %s, `type` = %s where `issue_id` = %s
        """, (buildtime, description, title, detected_type, issue_id))
        self._db.commit()
        return issue_id


    def add_to_issue(self, issue_id, rpms):
        if len(rpms) == 0: raise ValueError("no rpms given")
        query_values = []
        for rpm in rpms:
            rpm_data = rpminfo(rpm)
            query_values.append((
                issue_id, rpm_data['name'], rpm_data['epoch'], rpm_data['version'],
                rpm_data['release'], rpm_data['arch'], rpm_data['srpm'], rpm,
                rpm_data['checksum']
            ))
        try:
            rows = self.cursor.executemany("""
                insert into `packages` (`issue_id`, `name`, `epoch`, `version`,
                    `release`, `arch`, `srpm`, `filename`, `checksum`)
                    values (%s, %s, %s, %s, %s, %s, %s, %s, %s)
            """, query_values)
        except pymysql.err.IntegrityError:
            raise AssertionError("duplicate packages")
        if len(rpms) != rows: raise AssertionError(f"len(rpms) != {rows}")
        self._db.commit()
        return True


    def edit_issue(self, issue_ids, title=None, issue_type=None):
        modifier = []
        if title is not None: modifier.append(f"`title` = {title}")
        if issue_type is not None: modifier.append(f"`type` = {issue_type}")
        if len(modifier) == 0: return 0
        in_p = ', '.join(list(map(lambda x: str(x), issue_ids)))
        rows = self.cursor.execute(f"""
            update issues set %s where `issue_id` in ({in_p})
        """ % ",".join(modifier))
        self._db.commit()
        return rows


    def approve(self, issue_ids):
        in_p = ', '.join(list(map(lambda x: str(x), issue_ids)))
        rows = self.cursor.execute(f"""
            update issues set approved=1 where `issue_id` in ({in_p})
        """)
        self._db.commit()
        return rows


    def revoke(self, issue_ids):
        in_p = ', '.join(list(map(lambda x: str(x), issue_ids)))
        rows_pkgs = self.cursor.execute(f"""
            delete from packages where `issue_id` in ({in_p})
        """)
        rows_issues = self.cursor.execute(f"""
            update issues set description=\"***REVOKED***\",
            approved=0 where `issue_id`
            in ({in_p})
        """)
        self._db.commit()
        return rows_issues, rows_pkgs


    def purge(self, issue_ids):
        in_p = ', '.join(list(map(lambda x: str(x), issue_ids)))
        rows = self.cursor.execute(f"""
            delete from issues where `issue_id` in ({in_p}) and `approved` = 0
        """)
        self._db.commit()
        return rows


    def list(self):
        self.cursor.execute("""
            select `issue_id`, `issued_date`, `type`, `title`, `approved` from issues
        """)
        return self.cursor.fetchall()


    def show_issue(self, issue_ids=None, approved=None):
        query = """
            select `issue_id`, `issued_date`, `type`, `description`, `title`, `approved`
            from issues
        """
        if issue_ids is not None:
            in_p = ', '.join(list(map(lambda x: str(x), issue_ids)))
            query += f"where `issue_id` in ({in_p})"
        if approved is not None:
            if " where " not in query: query += " where "
            else: query += " and "
            query += f"approved = {int(approved)}"
        self.cursor.execute(query)
        return self.cursor.fetchall()


    def list_issue_packages(self, issue_id):
        self.cursor.execute("""
            select `name`, `epoch`, `version`, `release`, `arch`, `srpm`,
            `filename`, `checksum` from `packages` where `issue_id` = %s
        """ % issue_id)
        return self.cursor.fetchall()


    def last_issued(self):
        self.cursor.execute("""
            select unix_timestamp(max(issued_date)) as m from issues
        """)
        return int(self.cursor.fetchone()['m'])

    def get_meta(self, name, releasever):
        self.cursor.execute("""
            select short, description from meta
            where name = %s and releasever = %s
        """, (name, releasever))
        return self.cursor.fetchone()
